//NOTE This is a .h file but we expect it to be included exactly one (in main.cpp).

#include <stdint.h>
#include <bitset>
#include <Adafruit_MPR121.h>

constexpr uint8_t TOUCH_I2C_ADDRESS[2] = { 0x5a, 0x5c };

enum class TouchAreaType {
  NONE,
  SPOKE,
  INNER,
  MIDDLE,
  OUTER,
  SKREW,
  TODO
};

struct TouchAreaInfo {
  TouchAreaType type;

  // for SPOKE:
  //   0 is to the right (slightly too low)
  //   1 is above that (counter-clockwise)
  //   ...
  //   15 is the last one
  // for SKEW: 0..3
  // others: starting angle, like SPOKE
  uint8_t angle;
};

constexpr TouchAreaInfo TOUCH_AREAS[32] = {
  { TouchAreaType::SPOKE,   7 },  //  0
  { TouchAreaType::SPOKE,   6 },  //  1
  { TouchAreaType::SPOKE,   5 },  //  2
  { TouchAreaType::MIDDLE,  0 },  //  3
  { TouchAreaType::INNER,   0 },  //  4
  { TouchAreaType::SPOKE,   4 },  //  5
  { TouchAreaType::OUTER,   0 },  //  6
  { TouchAreaType::SPOKE,   2 },  //  7
  { TouchAreaType::SPOKE,   1 },  //  8
  { TouchAreaType::SPOKE,   0 },  //  9
  { TouchAreaType::SPOKE,   3 },  // 10,  OUTER?
  { TouchAreaType::SKREW,   2 },  // 11, upper right
  { TouchAreaType::NONE,    0 },  // 12
  { TouchAreaType::NONE,    0 },  // 13
  { TouchAreaType::NONE,    0 },  // 14
  { TouchAreaType::NONE,    0 },  // 15
  { TouchAreaType::SKREW,  13 },  // 16, lower right
  { TouchAreaType::SPOKE,  10 },  // 17
  { TouchAreaType::SPOKE,  14 },  // 18
  { TouchAreaType::SPOKE,  13 },  // 19
  { TouchAreaType::SPOKE,   9 },  // 20
  { TouchAreaType::SPOKE,  12 },  // 21
  { TouchAreaType::SPOKE,  11 },  // 22
  { TouchAreaType::MIDDLE,  8 },  // 23
  { TouchAreaType::OUTER,   8 },  // 24
  { TouchAreaType::INNER,   8 },  // 25
  { TouchAreaType::SPOKE,   8 },  // 26
  { TouchAreaType::SKREW,   5 },  // 27, upper left
  { TouchAreaType::NONE,    0 },  // 28
  { TouchAreaType::NONE,    0 },  // 29
  { TouchAreaType::NONE,    0 },  // 30
  { TouchAreaType::NONE,    0 },  // 31
};

struct HeizungTouch {
  Adafruit_MPR121 device[2];
  bool available[2];
  std::bitset<32> touchedHistory[4];

  void init() {
    for (int i=0; i<sizeof(touchedHistory)/sizeof(touchedHistory[0]); i++) {
      touchedHistory[i] = 0;
    }

    for (int i=0;i<2;i++) {
      available[i] = device[i].begin(TOUCH_I2C_ADDRESS[i]);
      if (!available[i])
        continue;
      auto& cap = device[i];

      // disable it, again
      cap.writeRegister(MPR121_ECR, 0x0);

      // auto-config code (mostly) copied from Adafruit library
      cap.writeRegister(MPR121_AUTOCONFIG0, 0x0B);

      // correct values for Vdd = 3.3V
      cap.writeRegister(MPR121_UPLIMIT, 200);     // ((Vdd - 0.7)/Vdd) * 256
      cap.writeRegister(MPR121_TARGETLIMIT, 180); // UPLIMIT * 0.9
      cap.writeRegister(MPR121_LOWLIMIT, 130);    // UPLIMIT * 0.65

      // enable X electrodes and start MPR121
      // enable virtual proximity electrode, as well
      cap.writeRegister(MPR121_ECR, 0x8c | 0x30);
    }
  }

  void update(unsigned long now) {
    for (int i=sizeof(touchedHistory)/sizeof(touchedHistory[0])-1; i>=1; i--) {
      touchedHistory[i] = touchedHistory[i-1];
    }

    uint16_t touched[2];
    for (int i=0;i<2;i++) {
      touched[i] = available[i] ? device[i].touched() : 0;
    }
    touchedHistory[0] = (touched[1] << 16) | touched[0];
  }

  bool hasChanged() {
    return (touchedHistory[0] != touchedHistory[1]);
  }

  bool hasChanged2() {
    return (touchedHistory[0] != touchedHistory[1]) || (touchedHistory[1] != touchedHistory[2]);
  }

  std::bitset<32> touched() {
    return touchedHistory[0];
  }

  std::bitset<32> touchedPrev() {
    return touchedHistory[1];
  }

  std::bitset<32> changed() {
    return touched() ^ touchedPrev();
  }

  std::bitset<32> pressed() {
    return touchedHistory[0] & touchedHistory[1] & ~touchedHistory[2] & ~touchedHistory[3];
  }

  std::bitset<32> released() {
    return ~touchedHistory[0] & ~touchedHistory[1] & touchedHistory[2] & touchedHistory[3];
  }

  std::bitset<32> pressedQuick() {
    return touchedHistory[0] & ~touchedHistory[1];
  }

  std::bitset<32> releasedQuick() {
    return ~touchedHistory[0] & touchedHistory[1];
  }
};
